/* Copyright (c) Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gtk/gtk.h>
#include <X11/Xatom.h>
#include <gdk/gdkx.h>
#include <vte/vte.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <mntent.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <libintl.h>
#include "splash.h"
#include "config.h"
#include "set-keymap.h"
#include <gdk-pixbuf/gdk-pixdata.h>
#include <libintl.h>
#include <popt.h> 
#include <sys/mount.h>

#ifndef MNT_DETACH
#define MNT_DETACH 0x00000002
#endif

#define TMPPATH "/etc/rhgb/temp"
#define SOCKET_NAME TMPPATH"/rhgb-socket"
#define DISPLAY_NO ":9"
#define LOCALHOST "127.0.0.1"
#define DISPLAY_STRING LOCALHOST DISPLAY_NO
#define DISPLAY_FNAME TMPPATH"/display"
#define PIPE_NAME TMPPATH"/rhgb-console"
#define X_CONFIG "/etc/rhgb/xorg.conf"
#define XLOGFILE TMPPATH"/XFree86.0.log"
#define X_NO_DRI "-extension XFree86-DRI"
#define X_COMMAND1 "/usr/bin/X11/X -s off -dpms -v -nolock -fp /usr/lib/X11/fonts/misc -logfile " XLOGFILE " " X_NO_DRI " " DISPLAY_NO " vt8"
#define X_COMMAND2 "/usr/bin/X11/X -config " X_CONFIG " -s off -dpms -v -nolock -fp /usr/lib/X11/fonts/misc -logfile " XLOGFILE " " X_NO_DRI " " DISPLAY_NO " vt8"
#define X_FAILSAFE "/usr/bin/X11/X -s off -dpms -v -nolock -fp /usr/lib/X11/fonts/misc -logfile " XLOGFILE " " DISPLAY_NO " vt8"
#define NETWORK_COMMAND "/sbin/ifconfig lo " LOCALHOST " up"
#define DISPLAY_ENV "DISPLAY=" DISPLAY_STRING
#define FIX_INIT_COMMAND "/sbin/change_console -r"
#define SYSRESOURCES "/etc/X11/Xresources"
#define XRDB "/usr/bin/X11/xrdb"
#define XRDB_COMMAND XRDB " -display " DISPLAY_STRING " -merge " SYSRESOURCES

/*
 * Maximum time to wait until the sysinit finished and we were asked to quit
 */
#define QUIT_TIMEOUT 15

/*
 * Time between we detected the main X server was started and we actually
 * switch console and exit. The main server detection is based on the
 * /tmp/.X0-lock file presence and X may is not graphically ready at that
 * point.
 */
#define X_TIMEOUT 3

/* Keep in sync w/ rhgb-client.c */
#define N_STEPS 21

/* For localizations with gettext */
#define _(x) dgettext (GETTEXT_PACKAGE, x)
#define N_(x) (x)


typedef struct _ServiceMapping ServiceMapping;
struct _ServiceMapping
{
  const char *service;
  const char *readable_name;
};

ServiceMapping service_list[] =
  {
    {"RCclock", N_("Setting up system clock")},
    {"RChostname", N_("Setting the hostname")},
    {"RCmountfs", N_("Setting up local disks")},
    {"random", N_("Starting random number generator")},
    {"gpm", N_("Starting console mouse services")},
    {"kudzu", N_("Probing for new hardware")},
    {"pcmcia", N_("Starting PCMCIA card services")},
    {"iptables", N_("Starting firewall")},
    {"ipchains", N_("Starting firewall")},
    {"apmd", N_("Starting power management services")},
    {"cups", N_("Starting printing system")},
    {"isdn", N_("Starting isdn")},
    {"network", N_("Starting networking")},
    {"sshd", N_("Starting sshd")},
    {"ntpd", N_("Starting time server")},
    {"rhnsd", N_("Starting Red Hat Network support")},
    {"sendmail", N_("Starting E-Mail transport")},
    {"spamassassin", N_("Starting E-Mail spam filter system")},
    {"firstboot", N_("Starting first-time boot configuration")},
    {"loginscreen", N_("Starting login services")}
  };

static gint n_intervals;
static gint current_interval = -1;
static gboolean force_arg = 0;
static gboolean debug = 0;
static gboolean debug_interface = 0;
static char *use_existing_server = NULL;

static gint child_pid = 0;
static GList *services = NULL;
static Splash *splash = NULL;
/* Timeouts */
static guint motion_timeout = 0;
static guint motion_notify_event_id = 0;

static gint socket_fd = 0;
static gint pipe_fd = 0;
static gint quit_timeout = 0;
static gint x_timeout = 0;


static const char *system_service_names[] =
{
  "RCconsole",
  "RCproc",
  "RCkernelparam",
  "RCclock",
  "RChostname",
  "RCusb",
  "RCmountfs",
  "RCswap",
  "RCraid"
};

static gboolean unmount_ramfs (void);

static void
DEBUG_MESSAGE (const char *message, ...)
{
  gchar *full_message;
  va_list args;

  if (!debug)
    return;

  va_start (args, message);
  full_message = g_strdup_vprintf (message, args);
  va_end (args);
  g_print (full_message);
  g_free (full_message);
}

/*
 * Test if the main graphical screen started
 * Returns TRUE if started and FALSE otherwise
 */
static gboolean
test_graphical(void) {
    struct stat buf;

    return (stat("/tmp/.X0-lock", &buf) == 0);
}

/*
 * The shutdown function. Kill the specific X process, and umount
 * the special filesystem
 */
static void
rhgb_shutdown(gint return_fd, int vt)
{
    if (child_pid > 0) {
        kill(child_pid, SIGKILL);
	if (vt > 0)
	    switch_to_vt(vt);
        waitpid(child_pid, NULL, 0);
	if (vt > 0)
	    switch_to_vt(vt);
    }
    if (return_fd >= 0) {
	write(return_fd, "done", strlen("done") + 1);
	close (return_fd);
    }
    close(socket_fd);
    close(pipe_fd);
    unlink(SOCKET_NAME);
    unlink(XLOGFILE);
    unlink(DISPLAY_FNAME);
    unlink(PIPE_NAME);
    unmount_ramfs();
    _exit(0);
}

/*
 * Timeout handling. If we don't get pings or progress informations
 * for a while we switch back automatically to text mode console
 * connections_timeout is reset to 0 each time we get some client command
 * If the boot process is stuck into some long command (fsck, network
 * or DNS start error) then it will automatically switch to the console mode
 * to show what is happening.
 */
static guint pipe_timeout = 0;
/*
 * If this is set to a negative number, we are supposed to
 * not timeout (e.g. for firstboot)
 */
static int connection_timeout = 0;
static int switched_vt1 = 0;
static int switched_details = 0;

static void connections_reset(int from_pipe) {
    pipe_timeout = 0;
    if (switched_vt1) {
        switch_to_vt(8);
	switched_vt1 = 0;
    }
    if (!from_pipe && connection_timeout >= 0)
	connection_timeout = 0;
    if ((!from_pipe) && (switched_details)) {
        /* switch back to no details */
	splash_hide_details(splash);
	switched_details = 0;
    }
}

static gboolean
connections_check(void) {
    if(connection_timeout >= 0) connection_timeout++;
    pipe_timeout++;
    /*
     * Those checks are intended to detect stops in progresses of the
     * boot and switch back/forth to/from the detailed or text views.
     */
    if ((connection_timeout >= 10) && (switched_vt1 == 0) &&
        (switched_details == 0)) {
        /*
	 * Something is blocking could be filesystem check or kudzu
	 * switch to detailed mode
	 */
        splash_show_details(splash);
	switched_details = 1;
    }
    if ((pipe_timeout >= 35) && (connection_timeout >= 35) &&
        (switched_vt1 == 0)) {
        /*
	 * something really bad is going on switch to text mode
	 */
        switch_to_vt(1);
	switched_vt1 = 1;
    }

    /*
     * those timeouts are needed in the last phase to detect X/gdm
     * actually started.
     */
    if (x_timeout > 0) {
        x_timeout--;
	connection_timeout = 0;
	if (x_timeout == 0) {
	    switch_to_vt(7);
	    rhgb_shutdown(-1, 7);
	}
    }
    if (quit_timeout > 0) {
        quit_timeout--;
	connection_timeout = 0;
	if (quit_timeout == 0) {
	    switch_to_vt(7);
	    rhgb_shutdown(-1, 7);
	} else if (test_graphical()) {
	    quit_timeout = 0;
	    x_timeout = X_TIMEOUT;
	}
    }
    return (TRUE);
}

static gboolean
init_timeout (void) {
    static int initialized = 0;

    if (initialized)
        return(FALSE);
    g_timeout_add(1000, (GSourceFunc) connections_check, NULL);
    initialized = 1;
    return(TRUE);
}

/* Check for x
 */
static gboolean
check_for_x (void)
{
  return (access ("/usr/bin/X11/X", R_OK | X_OK));
}

static gboolean
check_for_xrdb (void)
{
  if (!access(SYSRESOURCES, R_OK))
      return(FALSE);
  return (access (XRDB, R_OK | X_OK));
}

/* Get the language
 */
static gchar *
check_for_language (void)
{
#if (GLIB_MAJOR_VERSION > 2) || ((GLIB_MAJOR_VERSION == 2) && \
    (GLIB_MINOR_VERSION >= 4))
  gchar *contents, **words;
  gchar *ret = NULL;
  gsize len;
  int x;

  g_file_get_contents("/etc/sysconfig/keyboard",&contents,&len, NULL);
  if (!contents)
    return(NULL);

  words = g_strsplit_set(contents,"\n\r \t='\"_-",0);
  for (x = 0; words[x] ; x++)
    {
       if (!strcmp(words[x],"KEYTABLE"))
         {
	   x++;
	   while ((words[x] != NULL) && ((words[x][0] == 0))) x++;
	   ret = g_strdup(words[x]);
	   break;
	 }
    }
  g_free(contents);
  g_strfreev(words);
  return(ret);
#else
  return(NULL);
#endif
}

/* Check that we should start (i.e., that we are booting runlevel 5)
 */
static gboolean
check_for_run (void)
{
  gchar *contents, **args;
  gsize len;
  int x,rc = -1;
  gboolean rhgb_found = FALSE;
	

  /* Check that an alternate runlevel (other than 5) wasn't specified
   * on the command line */
  g_file_get_contents("/proc/cmdline",&contents,&len, NULL);
  if (!contents)
    return 1;
  contents = g_strstrip(contents);
  args = g_strsplit(contents," ",0);
  for (x = 0; args[x] ; x++)
    {
      if (g_ascii_isdigit(args[x][0]))
	{
	  if (atoi(args[x]) != 5)
	    rc = 1;
	  else
	    rc = 0;
	  break;
	}
      if (!strcmp(args[x],"-s") ||
	  !strcasecmp(args[x],"S") ||
	  !strcmp(args[x],"single"))
	{
	  rc = 1;
	  break;
	}
      if (!strcmp(args[x],"rhgb"))
	rhgb_found = TRUE;
    }
  g_free(contents);
  g_strfreev(args);
  if (rc >= 0) return rc;
  if (!rhgb_found) return 1;

  /* Check that the default runlevel is 5 */	
  g_file_get_contents("/etc/inittab",&contents,&len, NULL);
  if (!contents)
    return 1;
  args = g_strsplit(contents,"\n",0);
  for (x = 0; args[x] ; x++) {
    args[x] = g_strchug(args[x]);
    if (!strncmp(args[x],"id:",3))
      {
	if (atoi(args[x]+3) != 5)
	  rc = 1;
	else
	  rc = 0;
      }
  }
  g_free(contents);
  g_strfreev(args);
  return rc;
}

/*
 * Set up RAMfs for temporary files
 */
static gboolean
setup_ramfs(void)
{
    FILE *mtab;
    struct mntent *mountpoint;
    gboolean flag = FALSE;

    mtab = setmntent("/etc/mtab", "r");
    if (mtab == NULL)
        return 1;
    while ((mountpoint = getmntent(mtab)) != NULL) {
        if ((!strcmp(mountpoint->mnt_dir, TMPPATH)) &&
            (!strcmp(mountpoint->mnt_type, "ramfs"))) {
            flag = TRUE;
            break;
        }
    }
    endmntent(mtab);

    if (flag)
        return 0;

    if (mount("none", TMPPATH, "ramfs", 0, "maxsize=512"))
        return 1;

    chmod(TMPPATH, S_IRUSR | S_IWUSR | S_IXUSR);
    /* chdir(TMPPATH); */
    return 0;
}

/*
 * Unmount RAMfs
 */
static gboolean unmount_ramfs (void)
{
  return umount2(TMPPATH,MNT_DETACH);
}


/* Utility code
 */
static void
setup_failed (void)
{
  /* Do something */
  if (child_pid > 0)
    {
      kill (child_pid, SIGTERM);
      waitpid (child_pid, NULL, 0);
    }
  close (socket_fd);
  unmount_ramfs ();
  _exit (1);
}


/* Display
 */

static gboolean
display_init (void)
{
  gchar **argv;
  int fd;

  if (use_existing_server != NULL)
    {
      gchar *env;

      env = g_strdup_printf ("DISPLAY=%s", use_existing_server);
      putenv (env);
      return 0;
    }

  system (NETWORK_COMMAND);
  unsetenv ("XAUTHORITY");
  putenv ((char *) DISPLAY_ENV);
  
  fd = open(X_CONFIG, O_RDONLY);
  if (fd >= 0) {
      close(fd);
      g_shell_parse_argv (X_COMMAND2,
			  NULL, &argv,
			  NULL);
  } else {
      g_shell_parse_argv (X_COMMAND1,
			  NULL, &argv,
			  NULL);
  }

  g_spawn_async ("/", argv, NULL,
		 G_SPAWN_DO_NOT_REAP_CHILD,
		 NULL, NULL, &child_pid, NULL);
  g_strfreev (argv);
  fd = open(DISPLAY_FNAME, O_WRONLY | O_CREAT , 0644);
  if (fd >= 0) {
      write(fd, DISPLAY_STRING, strlen(DISPLAY_STRING));
      close(fd);
  }
  return 0;
}

static gboolean
display_init_failsafe (void)
{
  gchar **argv;
  int fd;

  if (use_existing_server != NULL)
    {
      gchar *env;

      env = g_strdup_printf ("DISPLAY=%s", use_existing_server);
      putenv (env);
      return 0;
    }

  unsetenv ("XAUTHORITY");
  putenv ((char *) DISPLAY_ENV);
  
  g_shell_parse_argv (X_FAILSAFE,
		      NULL, &argv,
		      NULL);

  g_spawn_async ("/", argv, NULL,
		 G_SPAWN_DO_NOT_REAP_CHILD,
		 NULL, NULL, &child_pid, NULL);
  g_strfreev (argv);
  fd = open(DISPLAY_FNAME, O_WRONLY | O_CREAT , 0644);
  if (fd >= 0) {
      write(fd, DISPLAY_STRING, strlen(DISPLAY_STRING));
      close(fd);
  }
  return 0;
}

static gboolean
merge_resources (void)
{
  system (XRDB_COMMAND);
  return 0;
}

static gboolean
hide_main_window_timeout_cb (gpointer data ATTRIBUTE_UNUSED)
{
  splash_hide_screen (splash);

  return FALSE;
}

static void
update_progress_bar (const char *step)
{
  GList *list;
  gint count = 0;
  gchar *text;
  guint i;

  for (list = services; list; list = list->next)
    {
      if (! strcmp (step, list->data))
	{
	  if (count > current_interval)
	    {
	      text = NULL;
	      for (i = 0; i < G_N_ELEMENTS (service_list); i++)
		{
		  if (!strcmp (step, service_list[i].service))
		    {
		      text = _(service_list[i].readable_name);
		    }
		  /* We know firstboot will be the last step, so we give it 6
		   * seconds to popup and then hide the window.  We also use a
		   * normal cursor.
		   * We also try to set the keyboard language since the value
		   * specified in the X config file is not taken into account
		   * on a read-only filesystem, the keymap is generated into
		   * file and then loaded by X, which fails when we start 
		   * rhgb, use setxkbmap to set it up.
		   */
		  if (!strcmp (step, "firstboot"))
		    {
		      gchar *lang;
		      GdkCursor *cursor;

                      connection_timeout = -1;
                      lang = check_for_language();
		      if (lang != NULL) {
		          gchar *argv[5] = {
			      (gchar *) "/usr/bin/X11/setxkbmap",
			      (gchar *) "-display",
			      (gchar *) DISPLAY_NO,
			      NULL,
			      NULL
			  };
			  argv[3] = lang;
			  g_spawn_async(NULL, &argv[0], NULL, 
			                G_SPAWN_DO_NOT_REAP_CHILD, NULL,
			                NULL, NULL, NULL);
		          g_free(lang);
		      }

		      /*
		       * then try more accurate but more likely to fail method
		       */
		      set_xkb_keymap();

		      g_timeout_add (6001, hide_main_window_timeout_cb, NULL);
		      if (motion_notify_event_id)
			{
			  motion_notify_event_id = 0;
			}
		      if (motion_timeout)
			{
			  g_source_remove (motion_timeout);
			  motion_timeout = 0;
			}

		      cursor = gdk_cursor_new (GDK_LEFT_PTR);
		      gdk_window_set_cursor (gdk_get_default_root_window (), cursor);
		      gdk_cursor_unref (cursor);
		    }
		}

	      current_interval = count;

	      if (n_intervals == 0)
		splash_pulse (splash);
	      else
		splash_set_progress (splash, (count+1)/(float)n_intervals);
	      /* We set the text after the progress so that the setting the
	       * progressbar doesn't clear it
	       */
	      if (text)
		{
		  splash_set_label (splash, text);
		}
	    }
	  break;
	}
      count++;
    }
}

/* Communication with the client
 */
static void
handle_command (char *command, gint return_fd)
{
  DEBUG_MESSAGE ("got a command:%s:\n", command);

  if (! strcmp (command, "quit"))
    {
      if (test_graphical()) {
          rhgb_shutdown(return_fd, 0);
      }
      else if (quit_timeout == 0) {
          quit_timeout = QUIT_TIMEOUT;
      }
      write(return_fd, "done", strlen("done") + 1);
      update_progress_bar("loginscreen");
    }
  else if (! strcmp (command, "ping"))
    {
      write (return_fd, "pong", strlen ("pong") + 1);
    }
  else if (! strncmp (command, "update=", strlen ("update=")))
    {
      update_progress_bar (command + strlen ("update=")); 
    }
  else if (!strncmp (command, "details", strlen ("details")))
    {
      splash_show_details (splash);
    }
  else if (!strncmp (command, "nodetails", strlen ("nodetails")))
    {
      splash_hide_details (splash);
    }
  else if (!strncmp (command, "sysinit", strlen ("sysinit")))
    {
      /* set the keyboard map */
      set_xkb_keymap ();
    }
  close (return_fd);
}

static gboolean
client_message_func (GIOChannel   *source,
		     GIOCondition  condition ATTRIBUTE_UNUSED,
		     gint          return_fd)
{
  gchar buffer[200];
  gsize bytes_read = 0;
  GIOStatus status;

  /* FIXME: We expect the command to be less than 200 bytes, come in one read,
   * and be NULL terminated.  We ignore everything that's not like that.
   */
  status = g_io_channel_read_chars (source,
				    buffer,
				    199,
				    &bytes_read,
				    NULL);

  switch (status)
    {
    case G_IO_STATUS_NORMAL:
      connections_reset(0);
      if (bytes_read >= 200)
	{
	  DEBUG_MESSAGE ("More than 200 bytes: %d\n", bytes_read);
	  goto done;
	}
      /* Make sure it's NULL terminated */
      buffer[bytes_read] = '\000';
      DEBUG_MESSAGE ("got the buffer:%d:%s:\n", bytes_read, buffer);
      handle_command (buffer, return_fd);
      goto done;
    case G_IO_STATUS_AGAIN:
      DEBUG_MESSAGE ("G_IO_STATUS_AGAIN\n");
      return TRUE;
    case G_IO_STATUS_EOF:
      DEBUG_MESSAGE ("G_IO_STATUS_EOF\n");
      break;
    case G_IO_STATUS_ERROR:
      DEBUG_MESSAGE ("G_IO_STATUS_ERROR\n");
      break;
    default:
      g_assert_not_reached ();
    }

 done:
  close (g_io_channel_unix_get_fd (source));
  return FALSE;
}


static gboolean
socket_func (GIOChannel   *source,
	      GIOCondition  condition,
	      gpointer      data ATTRIBUTE_UNUSED)
{
  if (condition == G_IO_IN)
    {
      GIOChannel *new_channel;
      gint new_fd;

      new_fd = accept (g_io_channel_unix_get_fd (source), NULL, 0);

      new_channel = g_io_channel_unix_new (new_fd);
      g_io_channel_set_encoding (new_channel, NULL, NULL);
      g_io_channel_set_flags (new_channel,
			      g_io_channel_get_flags (new_channel) | G_IO_FLAG_NONBLOCK, NULL);
      g_io_add_watch (new_channel, G_IO_IN, (GIOFunc) client_message_func, GINT_TO_POINTER (new_fd));
      g_io_channel_unref (new_channel);
    }
  return TRUE;
}

/*
 * create the socket to communicate with application rhgb_client
 * the socket is under SOCKET_NAME i.e. /etc/rhgb/temp/rhgb-socket
 */
static gboolean
socket_init (void)
{
  int curstate = 1;
  struct sockaddr_un *saddr; 
  GIOChannel *sock_channel;

  socket_fd = socket (PF_UNIX, SOCK_STREAM, 0);

  if (socket_fd < 0)
    return -1;

  if (setsockopt (socket_fd, SOL_SOCKET, SO_REUSEADDR, &curstate, sizeof(curstate)) < 0)
    return -1;

  /*
   * Make sure socket closed when exec commands happen
   */
  fcntl(socket_fd, F_SETFD, FD_CLOEXEC);

  saddr = g_new0 (struct sockaddr_un, 1);
  saddr->sun_family = AF_UNIX;
  memcpy (saddr->sun_path, SOCKET_NAME, sizeof (SOCKET_NAME));

  if (bind (socket_fd, (struct sockaddr *) saddr, sizeof (struct sockaddr_un)))
    return -1;

  if (listen (socket_fd, 5))
    return -1;

  chmod (SOCKET_NAME, S_IRUSR | S_IWUSR | S_IXUSR);

  sock_channel = g_io_channel_unix_new (socket_fd);
  g_io_add_watch (sock_channel, G_IO_IN, socket_func, NULL);
  g_io_channel_unref (sock_channel);

  return 0;
}

/*
 * create the fifo to echo data before init process console is reset
 * needed for rc.sysinit
 * the fifo is under PIPE_NAME i.e. /etc/rhgb/temp/rhgb-console
 */
static gboolean
pipe_func(GIOChannel * source,
          GIOCondition condition, gpointer data ATTRIBUTE_UNUSED)
{
    if (condition == G_IO_IN) {
        gchar buffer[4097];
        gsize bytes_read = 0;
        GIOStatus status;
        status = g_io_channel_read_chars(source,
                                         buffer, 4096, &bytes_read, NULL);

        switch (status) {
            case G_IO_STATUS_NORMAL:
                connections_reset(1);
                if (bytes_read > 4096) {
                    DEBUG_MESSAGE("More than 4096 bytes: %d\n", bytes_read);
                    goto done;
                }
                /* Make sure it's NULL terminated */
                buffer[bytes_read] = '\000';
                DEBUG_MESSAGE("got the pipe:%d:%s:\n", bytes_read,
                              buffer);
                splash_echo(splash, buffer, (glong) bytes_read);
                return TRUE;
            case G_IO_STATUS_AGAIN:
                DEBUG_MESSAGE("G_IO_STATUS_AGAIN\n");
                return TRUE;
            case G_IO_STATUS_EOF:
                DEBUG_MESSAGE("G_IO_STATUS_EOF\n");
                break;
            case G_IO_STATUS_ERROR:
                DEBUG_MESSAGE("G_IO_STATUS_ERROR\n");
                break;
            default:
                g_assert_not_reached();
        }

      done:
        close(g_io_channel_unix_get_fd(source));
	pipe_fd = -1;
        return FALSE;
    }
    return TRUE;
}

static gboolean
pipe_init (void)
{
    int ret;
    GIOChannel *pipe_channel;

    unlink(PIPE_NAME);

    ret = mkfifo(PIPE_NAME, S_IRUSR | S_IWUSR);
    if (ret != 0)
        return -1;

    pipe_fd = open(PIPE_NAME, O_NONBLOCK | O_RDWR);
    if (pipe_fd < 0)
        return -1;

    /*
     * Make sure the pipe is when exec commands happen
     */
    fcntl(pipe_fd, F_SETFD, FD_CLOEXEC);

    pipe_channel = g_io_channel_unix_new(pipe_fd);
    g_io_channel_set_encoding(pipe_channel, NULL, NULL);
    g_io_add_watch(pipe_channel, G_IO_IN, pipe_func, NULL);
    g_io_channel_unref(pipe_channel);
    return 0;
}


static const char *invalid_suffix[] = 
{
  ".rpmsave",
  ".rpmorig",
  ".rpmnew",
  "~",
  ".orig"
};

static gboolean
check_for_invalid_service (const char *name)
{
  guint i;
  gint name_len = strlen (name);

  g_assert (name);

  if (name[0] != 'S')
    return TRUE;

  if (name_len < 3)
    return TRUE;
  
  for (i = 0; i < G_N_ELEMENTS (invalid_suffix); i++)
    {
      const gchar *suffix;
      gint suffix_len;

      suffix = invalid_suffix[i];
      suffix_len = strlen (suffix);

      if (name_len <= suffix_len)
	continue;

      if (! strcmp (name + name_len - suffix_len, suffix))
	return TRUE;
    }
	  
  return FALSE;
}

static gboolean
init_list_of_services (void)
{
  struct dirent *dir;
  DIR *service_dir;
  gint i;
  GList *list;

  service_dir = opendir ("/etc/rc.d/rc5.d/");
  if (service_dir == NULL)
    return -1;
  while ((dir = readdir (service_dir)) != NULL)
    {
      if (check_for_invalid_service (dir->d_name))
	continue;

      services = g_list_prepend (services, g_strdup (dir->d_name));
    }
  closedir (service_dir);

  services = g_list_sort (services, (GCompareFunc)strcmp);

  /* now we stip off the leading 3 characters of each name */
  for (list = services; list; list = list->next)
    {
      gchar *service_name;

      service_name = g_strdup ((char *)list->data + 3);
      g_free (list->data);
      list->data = service_name;
    }

  for (i = G_N_ELEMENTS (system_service_names) - 1; i >= 0; i--)
    {
      services = g_list_prepend (services, g_strdup (system_service_names[i]));
    }

  n_intervals = g_list_length (services);

  return 0;
}


/* Display Window
 */

static void
create_main_window (int close_parent_fd)
{
  splash = splash_new ();

  close (close_parent_fd);
}


int
main (int argc, char *argv[])
{
  int pipe_fds[2] = { -1, -1 };
  int pid;
  int rc;
  poptContext context;

  const struct poptOption options[] = {
    POPT_AUTOHELP
    { "force", 'f', POPT_ARG_NONE, &force_arg, 1,
      N_("Force the graphical boot to start"), NULL },
    { "debug", 'd', POPT_ARG_NONE, &debug, 0,
      N_("leave debugging information up"), NULL },
    { "use-existing-server", 'u', POPT_ARG_STRING, &use_existing_server, 0,
      N_("use an existing server instead of creating one.  used primarily for debugging"), NULL },
    { "debug-interface", 'i', POPT_ARG_NONE, &debug_interface, 0,
      N_("Debug the graphical interface only"), NULL},
    { 0, 0, 0, 0, 0, 0, 0 }
  };

  bindtextdomain (GETTEXT_PACKAGE, "/usr/share/locale");
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
    
  context =  poptGetContext ("rhgb", argc, (const char **)argv,
			     options, 0);

  while ((rc = poptGetNextOpt (context)) > 0);
  if (rc < -1) {
    fprintf(stderr, "%s: %s\n",
	    poptBadOption(context, POPT_BADOPTION_NOALIAS),
	    poptStrerror(rc));
    return -1;
  }

  if (debug_interface)
    {
      gtk_init_check (&argc, &argv);
      create_main_window (pipe_fds[1]);
      gtk_main ();
      exit (0);
    }
  if (check_for_x () != 0)
    setup_failed ();
  
  if (check_for_run () != 0 && !force_arg)
    setup_failed ();

  if (pipe (pipe_fds) != 0)
    setup_failed ();

  if (setup_ramfs () != 0)
    setup_failed ();
  
  pid = fork ();
  if (pid < 0)
    setup_failed ();
  else if (pid > 0)
    {
      char *buf[5];
      close (pipe_fds[1]);
      read (pipe_fds[0], buf, 4);
      exit (0);
    }

  /* Set ourselves up as a daemon */
  close (pipe_fds[0]);

  if (! debug)
    {
      dup2 (open("/dev/null", O_RDWR), STDIN_FILENO);
      dup2 (open("/dev/null", O_RDWR), STDOUT_FILENO);
      dup2 (open("/dev/null", O_RDWR), STDERR_FILENO);
    }
  
  setsid ();
  setpgid (0, 0);

  if (display_init () != 0)
    setup_failed ();

  if (socket_init () != 0)
    setup_failed ();

  if (init_list_of_services () != 0)
    setup_failed ();

  if (gtk_init_check (&argc, &argv) == 0) {
      if (display_init_failsafe () != 0)
	  setup_failed ();

      if (gtk_init_check (&argc, &argv) == 0) {
	  setup_failed ();
      }
  }

  if (check_for_xrdb())
      merge_resources();

  create_main_window (pipe_fds[1]);

  if (pipe_init() != 0)
    setup_failed ();

  init_timeout();

  gtk_main ();

  system(FIX_INIT_COMMAND);

  return 0;
}
